from dataclasses import dataclass
from functools import lru_cache
from typing import List, Optional, Callable

from jose import jwt as jwt_utils

from .discovery import (
    get_discovery_document,
    DiscoveryDocumentRequest,
    DiscoveryDocumentResponse,
)
from .exceptions import PyIdentityModelException
from .jwks import get_jwks, JwksRequest, JsonWebKey, JwksResponse


@dataclass
class TokenValidationConfig:
    perform_disco: bool
    key: Optional[dict] = None
    audience: Optional[str] = None
    algorithms: Optional[List[str]] = None
    issuer: Optional[List[str]] = None
    subject: Optional[str] = None
    options: Optional[dict] = None
    claims_validator: Optional[Callable] = None


def _get_public_key(jwt: str, keys: List[JsonWebKey]) -> JsonWebKey:
    headers = jwt_utils.get_unverified_headers(jwt)
    filtered_keys = list(filter(lambda x: x.kid == headers["kid"], keys))
    if not filtered_keys:
        raise PyIdentityModelException("No matching kid found")

    key = filtered_keys[0]
    if not key.alg:
        key.alg = headers["alg"]

    return key


def _validate_token_config(
    token_validation_config: TokenValidationConfig,
) -> bool:
    if token_validation_config.perform_disco:
        return True

    if (
        not token_validation_config.key
        and not token_validation_config.algorithms
    ):
        raise PyIdentityModelException(
            "TokenValidationConfig.key and TokenValidationConfig.algorithms are required if perform_disco is False"
        )


@lru_cache
def _get_disco_response(disco_doc_address: str) -> DiscoveryDocumentResponse:
    return get_discovery_document(
        DiscoveryDocumentRequest(address=disco_doc_address)
    )


@lru_cache
def _get_jwks_response(jwks_uri: str) -> JwksResponse:
    return get_jwks(JwksRequest(address=jwks_uri))


def validate_token(
    jwt: str,
    token_validation_config: TokenValidationConfig,
    disco_doc_address: str = None,
) -> dict:
    _validate_token_config(token_validation_config)

    if token_validation_config.perform_disco:
        disco_doc_response = _get_disco_response(disco_doc_address)

        if not disco_doc_response.is_successful:
            raise PyIdentityModelException(disco_doc_response.error)

        jwks_response = _get_jwks_response(disco_doc_response.jwks_uri)
        if not jwks_response.is_successful:
            raise PyIdentityModelException(jwks_response.error)

        token_validation_config.key = _get_public_key(
            jwt, jwks_response.keys
        ).as_dict()

    decoded_token = jwt_utils.decode(
        jwt,
        token_validation_config.key,
        audience=token_validation_config.audience,
        algorithms=token_validation_config.algorithms,
        issuer=token_validation_config.issuer,
        subject=token_validation_config.subject,
        options=token_validation_config.options,
    )

    if token_validation_config.claims_validator:
        token_validation_config.claims_validator(decoded_token)

    return decoded_token


__all__ = ["validate_token", "TokenValidationConfig"]



import os

from dotenv import load_dotenv

load_dotenv()


def get_config() -> dict:
    return {
        "TEST_DISCO_ADDRESS": os.environ["TEST_DISCO_ADDRESS"],
        "TEST_JWKS_ADDRESS": os.environ["TEST_JWKS_ADDRESS"],
        "TEST_CLIENT_ID": os.environ["TEST_CLIENT_ID"],
        "TEST_CLIENT_SECRET": os.environ["TEST_CLIENT_SECRET"],
        "TEST_SCOPE": os.environ["TEST_SCOPE"],
        "TEST_EXPIRED_TOKEN": os.environ["TEST_EXPIRED_TOKEN"],
        "TEST_AUDIENCE": os.environ["TEST_AUDIENCE"],
    }



import os

from py_identity_model import JwksRequest, get_jwks
from .test_utils import get_config

TEST_JWKS_ADDRESS = get_config()["TEST_JWKS_ADDRESS"]


def test_get_jwks_is_successful():
    jwks_request = JwksRequest(address=TEST_JWKS_ADDRESS)
    jwks_response = get_jwks(jwks_request)
    assert jwks_response.is_successful
    for key in jwks_response.keys:
        assert key.kty
        assert key.alg
        assert key.use
        assert key.kid
        assert key.n
        assert key.e

        if key.x5t:
            assert key.x5c


def test_get_jwks_fails():
    jwks_request = JwksRequest(address="https://google.com")
    jwks_response = get_jwks(jwks_request)
    assert jwks_response.is_successful is False



from dataclasses import dataclass
from typing import Optional

import requests


@dataclass
class ClientCredentialsTokenRequest:
    address: str
    client_id: str
    client_secret: str
    scope: str


@dataclass
class ClientCredentialsTokenResponse:
    is_successful: bool
    token: Optional[dict] = None
    error: Optional[str] = None


def request_client_credentials_token(
    request: ClientCredentialsTokenRequest,
) -> ClientCredentialsTokenResponse:
    params = {"grant_type": "client_credentials", "scope": request.scope}

    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    response = requests.post(
        request.address,
        data=params,
        headers=headers,
        auth=(request.client_id, request.client_secret),
    )

    if response.ok:
        return ClientCredentialsTokenResponse(
            is_successful=True, token=response.json()
        )
    else:
        return ClientCredentialsTokenResponse(
            is_successful=False,
            error=f"Token generation request failed with status code: "
            f"{response.status_code}. Response Content: {response.content}",
        )


__all__ = [
    "ClientCredentialsTokenRequest",
    "ClientCredentialsTokenResponse",
    "request_client_credentials_token",
]



import os

from py_identity_model import (
    ClientCredentialsTokenRequest,
    request_client_credentials_token,
    get_discovery_document,
    DiscoveryDocumentRequest,
)
from .test_utils import get_config

config = get_config()
TEST_DISCO_ADDRESS = config["TEST_DISCO_ADDRESS"]
TEST_CLIENT_ID = config["TEST_CLIENT_ID"]
TEST_CLIENT_SECRET = config["TEST_CLIENT_SECRET"]
TEST_SCOPE = config["TEST_SCOPE"]


# TODO: failure conditions
def test_request_client_credentials_token_is_successful():
    disco_doc_response = get_discovery_document(
        DiscoveryDocumentRequest(address=TEST_DISCO_ADDRESS)
    )

    client_creds_req = ClientCredentialsTokenRequest(
        client_id=TEST_CLIENT_ID,
        client_secret=TEST_CLIENT_SECRET,
        address=disco_doc_response.token_endpoint,
        scope=TEST_SCOPE,
    )
    client_creds_token = request_client_credentials_token(client_creds_req)

    assert client_creds_token
    assert client_creds_token.is_successful
    assert client_creds_token.token


def test_request_client_credentials_token_fails():
    disco_doc_response = get_discovery_document(
        DiscoveryDocumentRequest(address=TEST_DISCO_ADDRESS)
    )

    client_creds_req = ClientCredentialsTokenRequest(
        client_id="bad_client_id",
        client_secret="bad_client_secret",
        address=disco_doc_response.token_endpoint,
        scope=TEST_SCOPE,
    )
    client_creds_token = request_client_credentials_token(client_creds_req)

    assert client_creds_token
    assert client_creds_token.is_successful is False
    print(client_creds_token.error)
    assert client_creds_token.error



from py_identity_model import DiscoveryDocumentRequest, get_discovery_document
from .test_utils import get_config

TEST_DISCO_ADDRESS = get_config()["TEST_DISCO_ADDRESS"]


def test_get_discovery_document_is_successful():
    disco_doc_request = DiscoveryDocumentRequest(address=TEST_DISCO_ADDRESS)
    disco_doc_response = get_discovery_document(disco_doc_request)
    assert disco_doc_response.is_successful
    assert disco_doc_response.issuer
    assert disco_doc_response.jwks_uri
    assert disco_doc_response.token_endpoint
    assert disco_doc_response.authorization_endpoint


def test_get_discovery_document_fails():
    disco_doc_request = DiscoveryDocumentRequest(address="https://google.com")
    disco_doc_response = get_discovery_document(disco_doc_request)
    assert disco_doc_response.is_successful is False



from dataclasses import dataclass
from typing import List, Optional

import requests


@dataclass
class JwksRequest:
    address: str


@dataclass
class JsonWebKey:
    kty: str
    use: str
    kid: str
    n: str
    e: str
    x5t: str = None
    x5c: List[str] = None
    issuer: Optional[str] = None
    alg: Optional[str] = None

    def as_dict(self):
        return {
            "kty": self.kty,
            "use": self.use,
            "kid": self.kid,
            "x5t": self.x5t,
            "n": self.n,
            "e": self.e,
            "x5c": self.x5c,
            "issuer": self.issuer,
            "alg": self.alg,
        }


@dataclass
class JwksResponse:
    is_successful: bool
    keys: Optional[List[JsonWebKey]] = None
    error: Optional[str] = None


def get_jwks(jwks_request: JwksRequest) -> JwksResponse:
    response = requests.get(jwks_request.address)
    # TODO: raise for status and handle exceptions
    if response.ok and "application/json" in response.headers.get(
        "Content-Type", ""
    ):
        response_json = response.json()
        keys = [JsonWebKey(**key) for key in response_json["keys"]]
        return JwksResponse(is_successful=True, keys=keys)
    else:
        return JwksResponse(
            is_successful=False,
            error=f"JSON web keys request failed with status code: "
            f"{response.status_code}. Response Content: {response.content}",
        )


__all__ = ["JwksRequest", "JwksResponse", "JsonWebKey", "get_jwks"]



from dataclasses import dataclass
from typing import Optional

import requests


@dataclass
class DiscoveryDocumentRequest:
    address: str


# TODO: full disco doc support
@dataclass
class DiscoveryDocumentResponse:
    is_successful: bool
    issuer: Optional[str] = None
    jwks_uri: Optional[str] = None
    authorization_endpoint: Optional[str] = None
    token_endpoint: Optional[str] = None
    error: Optional[str] = None


def get_discovery_document(
    disco_doc_req: DiscoveryDocumentRequest,
) -> DiscoveryDocumentResponse:
    response = requests.get(disco_doc_req.address)
    # TODO: raise for status and handle exceptions
    if response.ok and "application/json" in response.headers.get(
        "Content-Type", ""
    ):
        response_json = response.json()
        return DiscoveryDocumentResponse(
            issuer=response_json["issuer"],
            jwks_uri=response_json["jwks_uri"],
            authorization_endpoint=response_json["authorization_endpoint"],
            token_endpoint=response_json["token_endpoint"],
            is_successful=True,
        )
    else:
        return DiscoveryDocumentResponse(
            is_successful=False,
            error=f"Discovery document request failed with status code: "
            f"{response.status_code}. Response Content: {response.content}",
        )


__all__ = [
    "DiscoveryDocumentRequest",
    "DiscoveryDocumentResponse",
    "get_discovery_document",
]



import datetime

import pytest
from jose import ExpiredSignatureError

from py_identity_model import (
    PyIdentityModelException,
    validate_token,
    get_discovery_document,
    DiscoveryDocumentRequest,
    ClientCredentialsTokenRequest,
    request_client_credentials_token,
    TokenValidationConfig,
)
from py_identity_model.token_validation import (
    _get_disco_response,
    _get_jwks_response,
)
from .test_utils import get_config

# TODO: clean this up
config = get_config()
TEST_DISCO_ADDRESS = config["TEST_DISCO_ADDRESS"]
TEST_EXPIRED_TOKEN = config["TEST_EXPIRED_TOKEN"]
TEST_CLIENT_ID = config["TEST_CLIENT_ID"]
TEST_CLIENT_SECRET = config["TEST_CLIENT_SECRET"]
TEST_SCOPE = config["TEST_SCOPE"]
TEST_AUDIENCE = config["TEST_AUDIENCE"]

DEFAULT_OPTIONS = {
    "verify_signature": True,
    "verify_aud": False,
    "verify_iat": True,
    "verify_exp": True,
    "verify_nbf": True,
    "verify_iss": True,
    "verify_sub": True,
    "verify_jti": True,
    "verify_at_hash": True,
    "require_aud": False,
    "require_iat": False,
    "require_exp": False,
    "require_nbf": False,
    "require_iss": False,
    "require_sub": False,
    "require_jti": False,
    "require_at_hash": False,
    "leeway": 0,
}


def _generate_token():
    disco_doc_response = get_discovery_document(
        DiscoveryDocumentRequest(address=TEST_DISCO_ADDRESS)
    )
    client_creds_req = ClientCredentialsTokenRequest(
        client_id=TEST_CLIENT_ID,
        client_secret=TEST_CLIENT_SECRET,
        address=disco_doc_response.token_endpoint,
        scope=TEST_SCOPE,
    )
    client_creds_response = request_client_credentials_token(client_creds_req)
    return client_creds_response


def test_token_validation_expired_token():
    with pytest.raises(ExpiredSignatureError):
        validate_token(
            jwt=TEST_EXPIRED_TOKEN,
            disco_doc_address=TEST_DISCO_ADDRESS,
            token_validation_config=TokenValidationConfig(
                perform_disco=True, options=DEFAULT_OPTIONS
            ),
        )


def test_token_validation_succeeds():
    client_creds_response = _generate_token()

    validation_config = TokenValidationConfig(
        perform_disco=True, audience=TEST_AUDIENCE, options=DEFAULT_OPTIONS
    )

    claims = validate_token(
        jwt=client_creds_response.token["access_token"],
        disco_doc_address=TEST_DISCO_ADDRESS,
        token_validation_config=validation_config,
    )
    assert claims
    assert claims["iss"]
    assert claims["iat"]
    assert claims["exp"]


def test_token_validation_with_invalid_config_throws_exception():
    client_creds_response = _generate_token()

    validation_config = TokenValidationConfig(
        perform_disco=False, audience=TEST_AUDIENCE, options=DEFAULT_OPTIONS
    )

    with pytest.raises(PyIdentityModelException):
        validate_token(
            jwt=client_creds_response.token["access_token"],
            disco_doc_address=TEST_DISCO_ADDRESS,
            token_validation_config=validation_config,
        )


def test_cache_succeeds():
    client_creds_response = _generate_token()

    validation_config = TokenValidationConfig(
        perform_disco=True, audience=TEST_AUDIENCE, options=DEFAULT_OPTIONS
    )

    for i in range(0, 5):
        validate_token(
            jwt=client_creds_response.token["access_token"],
            disco_doc_address=TEST_DISCO_ADDRESS,
            token_validation_config=validation_config,
        )

    cache_info = _get_disco_response.cache_info()
    print(cache_info)
    assert cache_info
    assert cache_info[0] > 0

    cache_info = _get_jwks_response.cache_info()
    print(cache_info)
    assert cache_info
    assert cache_info[0] > 0


def test_benchmark_validation():
    client_creds_response = _generate_token()
    validation_config = TokenValidationConfig(
        perform_disco=True, audience=TEST_AUDIENCE, options=DEFAULT_OPTIONS
    )
    start_time = datetime.datetime.now()

    for i in range(0, 100):
        validate_token(
            jwt=client_creds_response.token["access_token"],
            disco_doc_address=TEST_DISCO_ADDRESS,
            token_validation_config=validation_config,
        )
    elapsed_time = datetime.datetime.now() - start_time
    print(elapsed_time)
    assert elapsed_time.total_seconds() < 1


def test_claim_validation_function_succeeds():
    client_creds_response = _generate_token()

    def validate_claims(token: dict):
        # Do some token validation here
        # and raise an exception if the validation fails
        pass

    validation_config = TokenValidationConfig(
        perform_disco=True,
        audience=TEST_AUDIENCE,
        options=DEFAULT_OPTIONS,
        claims_validator=validate_claims,
    )

    decoded_token = validate_token(
        jwt=client_creds_response.token["access_token"],
        disco_doc_address=TEST_DISCO_ADDRESS,
        token_validation_config=validation_config,
    )

    assert decoded_token
    assert decoded_token["iss"]


def test_claim_validation_function_fails():
    client_creds_response = _generate_token()

    def validate_claims(token: dict):
        raise PyIdentityModelException("Validation failed!")

    validation_config = TokenValidationConfig(
        perform_disco=True,
        audience=TEST_AUDIENCE,
        options=DEFAULT_OPTIONS,
        claims_validator=validate_claims,
    )

    with pytest.raises(PyIdentityModelException):
        validate_token(
            jwt=client_creds_response.token["access_token"],
            disco_doc_address=TEST_DISCO_ADDRESS,
            token_validation_config=validation_config,
        )



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = ["python-jose", "requests"]

test_require = ["python-jose", "requests", "pytest"]

setup(
    name="py_identity_model",
    version="0.11.5",
    description="OAuth2.0 and OpenID Connect Client Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jamescrowley321/py-oidc",
    author="James Crowley",
    author_email="jamescrowley151@gmail.com",
    license="Apache 2.0",
    platforms="Any",
    install_requires=install_requires,
    extras_require={"test": test_require},
    keywords="OpenID jwt",
    packages=["py_identity_model"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
)
